MAX = +1
MIN = -1

LinearMinMax_ = "LinearMinMax"
LinearMax_ = "LinearMax"
LinearSum_ = "LinearSum"
Vector_ = "Vector"
EnhancedAccuracy_ = "EnhancedAccuracy"
Logarithmic_ = "Logarithmic"

NORMALIZATION_METHODS = [LinearMinMax_,
                         LinearMax_,
                         LinearSum_,
                         Vector_,
                         EnhancedAccuracy_,
                         Logarithmic_]
                         
SAATY_SCALE = [1, 2 ,3 ,4 ,5 ,6 ,7 ,8 ,9, 1/2, 1/3, 1/4, 1/5, 1/6, 1/7, 1/8, 1/9]

SAATY_RI = [0, 0, 0.58, 0.9, 1.12, 1.24, 1.32, 1.41, 1.45, 1.49, 1.51, 1.53, 1.56, 1.57, 1.59]